/* Copyright (C) 2011 Papavasileiou Dimitris                             
 *                                                                      
 * This program is free software: you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation, either version 3 of the License, or    
 * (at your option) any later version.                                  
 *                                                                      
 * This program is distributed in the hope that it will be useful,      
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        
 * GNU General Public License for more details.                         
 *                                                                      
 * You should have received a copy of the GNU General Public License    
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __CONFIG_H_
#define __CONFIG_H_

/* This enables a work-around in the tachometer speed calculation code
   that seems to be necessary with my ECU.  The macro is named as it
   is under the assumption that the ECU is in fact designed to behave
   like this so that all NC29s will need this switch.  If you can
   confirm whether this is true or not please drop me a line so that I
   can rename the macro to something more suitable like
   MY_BROKEN_CDI_HACK. */

#define CBR_400_RR_NC29

/****
  ENGINE_ALERT_SPEED is the engine speed in rpms after which the
  tachometer display will begin to flash.

  SAMPLING_WINDOW_SIZE is the sliding window (in samples) used to
  smooth out all engine-related measurements.  This includes the
  coolant temperature, the battery voltage as well as engine
  speed. Bigger windows will result in a smoother reading but also in
  more sluggish response.
****/

#define ENGINE_ALERT_SPEED 13000
#define SAMPLING_WINDOW_SIZE 12
#define PULSES_PER_REVOLUTION 2

/****
  The circuit used to sample the coolant temperature sensor looks
  like this:
_______                         ________
       |    COOLANT_RESISTOR   |        |
  GPIO |=---/\/\/\/\/\/\/\/\+-=|-/\/\/\-|=----GND
       |                    |  |________|
  ADC  |=-------------------+    SENSOR
       |
   
  COOLANT_PORT, COOLANT_DDR and COOLANT_BIT specify the port and pin
  that is used to power the resistive sensor.

  COOLANT_CHANNEL specifies the ADC channel connected to the voltage
  divider.  It corresponds to bits MUX4..0 of the ADMUX register so that
  values 0-7 select the single-ended channels while greater values
  select differential channels with gain.

  COOLANT_VREF is the reference voltage of the ADC when measuring the
  coolant temperature sensor which is AVcc so this should normally be
  either 5 or 3.3 depending on how you're powering the uC.

  COOLANT_GAIN is the gain of the ADC if a differential channel is
  selected set this to 1 when using a single-ended channel or a
  differential channel with no gain.

  COOLANT_RESISTOR is the resistance of the resistor that is being used to
  form the voltage divider in Ohms.

  COOLANT_THRESHOLD is the measured sensor resistance in Ohms above which no
  temperature value will be shown.
****/

#define COOLANT_PORT PORTD
#define COOLANT_DDR DDRD
#define COOLANT_BIT 5
#define COOLANT_CHANNEL 13
#define COOLANT_VREF 5         /* Volts */
#define COOLANT_GAIN 10
#define COOLANT_RESISTOR 1800  /* Ohms */
#define COOLANT_THRESHOLD 165  /* Ohms */

/****
  Temperature sensor mapping
  --------------------------
  
  T: 44    48  68 76 80 85 92 100  106  Degrees
  R: 154.5 139 68 48 42 38 34 27.5 22.5 Ohms

  This data has been fitted to a 4th-degree polynomial so that
  T = COOLANT_COEFF_0 + COOLANT_COEFF_n * R^n, for n = [1, 4]
****/

#define COOLANT_COEFF_4 7.4481e-07
#define COOLANT_COEFF_3 -3.2525e-04
#define COOLANT_COEFF_2 5.1125e-02
#define COOLANT_COEFF_1 -3.6836e+00
#define COOLANT_COEFF_0 1.6780e+02

/****
  The circuit used to sample the battery voltage looks like this:
_______                       
       |           VOLTAGE_R1    VOLTAGE_R2
       |   Vbat----/\/\/\/\/\--+--/\/\/\/\/\----GND
       |                       |
  ADC  |=----------------------+
       |

  VOLTAGE_CHANNEL specifies the ADC channel connected to the voltage
  divider.  It corresponds to bits MUX4..0 of the ADMUX register so
  that values 0-7 select the single-ended channels while greater
  values select differential channels with gain.

  VOLTAGE_VREF is the reference voltage of the ADC when measuring the
  coolant temperature sensor which is AVcc so this should normally be
  either 5 or 3.3 depending on how you're powering the uC.

  VOLTAGE_R1, VOLTAGE_R2 are the resistances of the resistors forming
  the voltage divider.  The units are arbitrary provided that they're
  the same for both R1 and R2.
****/

#define VOLTAGE_CHANNEL 0
#define VOLTAGE_VREF 5       /* Volts */
#define VOLTAGE_R1 26.4      /* Kilo-Ohms */
#define VOLTAGE_R2 11.84     /* Kilo-Ohms */

/* These micros specify the port and pin connected to the ambient
   temperature thermometer's data line. */

#define THERMOMETER_PORT PORTE
#define THERMOMETER_DDR  DDRE
#define THERMOMETER_PIN  PINE
#define THERMOMETER_BIT  1

/****
  The macros below specify how the display is connected to the micro-
  controller.

  The CONTROL_* macros specify the port as well as the specific pin of
  the port that each of the control lines of the GLCD is connected to.

  The DATA_* macros specify the port connected to the data lines.  It
  is assumed that pin 0 of the port is connect the first data bit, pin
  1 to the second and so forth.
****/

#define CONTROL_PORT PORTB
#define CONTROL_DDR DDRB

#define CONTROL_DI_BIT    0
#define CONTROL_RW_BIT    1
#define CONTROL_EN_BIT    2
#define CONTROL_CS1_BIT   3
#define CONTROL_CS2_BIT   4
#define CONTROL_CS3_BIT   5
#define CONTROL_RESET_BIT 6

#define DATA_PORT PORTC
#define DATA_DDR DDRC
#define DATA_PIN PINC

/* The macros below specify the port and pin that is used to switch the
   ground when dimming the indicator leds.  It should be set to the pin
   assigned to the OC0A function for the microcontroller used. */

#define INDICATOR_DDR DDRB
#define INDICATOR_BIT 7

/***
  Here's the circuit used to attach the buttons to the microcontroller:

_______                       
       |      
    P0 |=----=^=---+
    P1 |=----=^=---+
    P2 |=----=^=---+
    P3 |=----=^=---+
    P4 |=----=^=---+
    P5 |=----=^=---+
    P6 |=----------+

  The BUTTONS_GROUND_ON_PIN macro should be one if the circuit above
  is used. In this case P6 (which is specified by BUTTONS_GROUND_BIT)
  is configured as an output and pulled low thus pulling the right
  side of all buttons to ground.  If BUTTONS_GROUND_ON_PIN is set to
  zero the right side of all buttons should be wired to a ground point
  instead.  The option to use a microcontroller pin as a ground point
  exists merely as a convenience so that a cable ribbon can be used to
  wire all buttons to the same port.

  The BUTTONS_*_BIT macros specify the pin that each of the six
  buttons is connected to.  In the current implementation not all six
  buttons are necessary to navigate the displays.  Buttons D, E and F
  are used as convenience shortcuts.

  In fact no buttons are strictly necessary.  The device can be
  controlled through a single switch.  This is assumed to be some sort
  of signal line that is disconnected or floating when the switch is
  not pressed and connected to battery voltage (or in fact any voltage
  sufficiently higher that Vcc) otherwise.  This has been designed
  with the vehicle signal switch in mind.  By connecting the wire
  leading to the long beam indicator light to the pin specified by the
  SIGNAL_* macros, which should be an external interrupt pin, the
  signal or passing switch can be used to control the device.  This
  has the annoying side-effect of signalling the oncoming traffic
  (unless suitable modifications to the vehicle's wiring are
  performed) but is nevertheless simple to connect and can be very
  useful, especially on motorcycles.
***/

#define BUTTONS_DDR DDRA
#define BUTTONS_PORT PORTA
#define BUTTONS_PIN PINA

#define BUTTONS_GROUND_ON_PIN 1
#define BUTTONS_GROUND_BIT 7

#define BUTTONS_A_BIT 2
#define BUTTONS_B_BIT 6
#define BUTTONS_C_BIT 1
#define BUTTONS_D_BIT 5
#define BUTTONS_E_BIT 0
#define BUTTONS_F_BIT 4

#define SIGNAL_INTERRUPT INT1
#define SIGNAL_PIN PIND
#define SIGNAL_BIT 1

/****
  These macros specify the time-intervals used when sampling any
  button.  All intervals are givec in milliseconds.

  DEBOUNCE_TIME is the time needed for the switch to be considered
  stable or debounced.

  INTERCLICK_TIME is the maximum time allowed between two consecutive
  clicks if they're to be considered as one double-click.
  
  LONG_CLICK_MIN_TIME and LONG_CLICK_MAX_TIME is the minimum and
  maximum time a button needs to be held down in order for the click
  to be read as a "long click".  If the maximum time is exceeded no
  click is registered at all.  This can be annoying but is needed for
  the long-beam signal interface to be usable since turning on the
  high beam would otherwise be registered as a long click.
****/

#define DEBOUNCE_TIME (20)
#define INTERCLICK_TIME (250)
#define LONG_CLICK_MIN_TIME (1000)
#define LONG_CLICK_MAX_TIME (3000)

/****
  Don't sample linear and lateral accelerations when below this
  speed threshold.  This is useful if your gps receiver gives
  bad speed and/or orientation readings for slow speeds.
****/
#define ACCELERATION_SAMPLING_THRESHOLD (20)

#endif
