/* Copyright (C) 2011 Papavasileiou Dimitris                             
 *                                                                      
 * This program is free software: you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation, either version 3 of the License, or    
 * (at your option) any later version.                                  
 *                                                                      
 * This program is distributed in the hope that it will be useful,      
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        
 * GNU General Public License for more details.                         
 *                                                                      
 * You should have received a copy of the GNU General Public License    
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <stdlib.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <math.h>

#include "tiny.h"
#include "large.h"
#include "state.h"
#include "common.h"
#include "config.h"

#define USART_BAUDRATE 38400
#define BAUD_PRESCALE (((F_CPU / (USART_BAUDRATE * 16UL))) - 1)

static void execute (unsigned char chip,
		     unsigned char flags,
		     unsigned char data)
{
    chip = (~(1 << (chip + 2))) & ((1 << CONTROL_CS1_BIT) |
				   (1 << CONTROL_CS2_BIT) |
				   (1 << CONTROL_CS3_BIT));
    
    /* Setup port C as an input port
       and poll display's status until
       it is ready to accept a command. */
    
    DATA_DDR = 0;

    do {
    	CONTROL_PORT = (1 << CONTROL_EN_BIT) |
	               (1 << CONTROL_RW_BIT) |
	               (1 << CONTROL_RESET_BIT) | chip;
    	CONTROL_PORT &= ~(1 << CONTROL_EN_BIT);
    } while ((DATA_PIN >> 7) || ((DATA_PIN >> 4) & 1));

    /* Reset port C as an output and
       execute the instruction. */

    CONTROL_PORT = chip | flags |
	           (1 << CONTROL_RESET_BIT) |
	           (1 << CONTROL_EN_BIT);
    
    DATA_DDR = ~0;    
    DATA_PORT = data;
    
    CONTROL_PORT &= ~(1 << CONTROL_EN_BIT);
}

static void set_address (unsigned char chip, unsigned char x, unsigned char y)
{
    execute (chip, 0, 0b10111000 | x);    
    execute (chip, 0, 0b01000000 | y);
}

static void write_column (unsigned char chip, unsigned char data)
{
    execute (chip, (1 << CONTROL_DI_BIT), data);
}

static int measure_tiny (char *string, int spacing)
{
    int i, j;

    for (i = 0, j = 0;
	 i < strlen (string);
	 j += (int)pgm_read_byte(&(tiny_widths[string[i] - ' '])) + spacing,
	 i += 1);

    return j;
}

static int measure_large (char *string, int spacing)
{
    int i, j;

    for (i = 0, j = 0;
	 i < strlen (string);
	 j += 20 + spacing,
	 i += 1);

    return j;
}

static void clear_tiny (unsigned char line, unsigned char column,
			unsigned char width)
{
    int i, k;
    
    if (column > 191) {
	return;
    }
	
    set_address ((column >> 6) + 1, line, column & 63);
	
    /* Blank out the given field. */
	
    for (i = 0, k = column ; i < width ; i += 1, k += 1) {
	if (k > 191) {
	    return;
	} else if (k == 64 || k == 128) {
	    set_address ((k >> 6) + 1, line, k & 63);
	}
	    
	write_column ((k >> 6) + 1, 0);
    }
}

static void put_tiny (unsigned char line, unsigned char column,
		      char *string, int spacing, int delta, char blink)
{
    int i, j, k, l, g;

    for (l = 0 ; l < (delta > 0 ? 2 : 1) ; l += 1) {
	for (j = 0, k = column ; j < strlen (string) ; j += 1) {
	    PGM_P glyph;
	    unsigned char width;
	    
	    if (k > 191) {
		return;
	    }
	
	    set_address ((k >> 6) + 1, line + l, k & 63);

	    glyph = (PGM_P)pgm_read_word (&(tiny_glyphs[string[j] - ' ']));
	    width = (unsigned char)pgm_read_byte(&(tiny_widths[string[j] - ' ']));
	
	    /* Write the glyph + spacing. */
	
	    for (i = 0 ; i < width + spacing ; i += 1, k += 1) {
		if (k > 191) {
		    return;
		} else if (k == 64 || k == 128) {
		    set_address ((k >> 6) + 1, line + l, k & 63);
		}

		g = l > 0 ?
		    pgm_read_byte(&(glyph[i])) >> (8 - delta) :
		    pgm_read_byte(&(glyph[i])) << delta;

		write_column ((k >> 6) + 1,
			      i < width && (timing.blink || (j != blink)) ?
			      g : 0);
	    }
	}
    }
}

static void put_large (unsigned char line, unsigned char column,
		       char *string, int spacing)
{
    int i, j, k, l;

    for (l = 0 ; l < 4 ; l += 1) {
	for (j = 0, k = column ; j < strlen (string) ; j += 1) {
	    PGM_P glyph;

	    if (k > 191) {
		return;
	    }

	    set_address ((k >> 6) + 1, line + l, k & 63);

	    glyph = (PGM_P)pgm_read_word (&(large_glyphs[string[j] - ' ']));

	    for (i = 0 ; i < 20 + spacing ; i += 1, k += 1) {
		if (k > 191) {
		    return;
		} else if (k == 64 || k == 128) {
		    set_address ((k >> 6) + 1, line + l, k & 63);
		}
	    
		write_column ((k >> 6) + 1,
			      i < 20 ? pgm_read_byte(&(glyph[(i * 4) + 3 - l])) : 0);
	    }
	}
    }
}

static void put_bar (unsigned char line, unsigned char column,
		     unsigned char width, signed char height,
		     signed char slope, unsigned char set)
{
    int i, k, l, h;
    int pattern;

    if (height > 0) {
	pattern = ~0;
    } else {
	pattern = 0b10101010;
	height = abs(height);
    }
    
    for (l = 0, k = column ; height > 0 ; l += 1, height = height > 8 ? height - 8 : 0) {
	if (k > 191) {
	    return;
	}

	set_address ((k >> 6) + 1, line - l, k & 63);
	
	for (i = 0 ; i < width ; i += 1) {
	    if (k + i > 191) {
		return;
	    } else if (k + i == 64 || k + i == 128) {
		set_address (((k + i) >> 6) + 1, line - l, (k + i) & 63);
	    }

	    if (slope > 0) {
		h = i < width >> 1 ? height - 1 : height;
	    } else if (slope < 0) {
		h = i > width >> 1 ? height - 1 : height;
	    } else {
		h = height;
	    }
	    
	    if (set) {
		unsigned char column;

		column = (h >= 8 ?
			  ~0 :
			  (unsigned char)(~0 - (1 << (8 - h)) + 1));
		column &= (k % 2 || pattern == ~0) ? pattern : ~pattern;

		if (h <= 8) {
		    column |= 1 << (8 - h);
		}
	    
		write_column (((k + i) >> 6) + 1, column);
	    } else {
		write_column (((k + i) >> 6) + 1, 0);
	    }
	}
    }
}

static void draw_bar_graph (void)
{
    signed char slopes[] = {2, 2, 1, 2, 0, 0, 1, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0,
			    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			    0, 0, 0, -2, 0, 0, 0, -1, 0, 0, -2, -1, -2, -2};
    
    unsigned char heights[] = {0, 11,
			       1, 12,
			       2, 13, 
			       3, 14,
			       6, 15,
			       7, 15,
			       8, 15,
			       10, 16,
			       38, 15,
			       40, 15,
			       42, 14,
			       44, 14,
			       45, 13,
			       46, 12,
			       47, 11,
			       100, 11};
    
    int i, j, k;

    for (i = 0 ; i < 48 ; i += 1) {
    	for (j = 0 ; i >= heights[j + 2] ; j += 2);

    	if (i < 12) {
    	    k = i % 5 < 2 ? 6 : 7;
    	} else if (i > 15) {
    	    k = (i - 16) % 6 < 2 ? 6 : 7;
    	} else {
    	    k = 7;
    	}

    	put_bar (k, 4 * i, 3, heights[j + 1] - 8 * (7 - k), slopes[i], (i < engine.speed / 334 && (engine.speed < ENGINE_ALERT_SPEED || timing.subblink)));
    }

    put_tiny (7, 0, "\x82", 1, 0, -1);
    put_tiny (7, 20, "\x83", 1, 0, -1);
    put_tiny (7, 40, "\x84", 1, 0, -1);
    put_tiny (7, 64, "\x85", 1, 0, -1);
    put_tiny (7, 88, "\x86", 1, 0, -1);
    put_tiny (7, 112, "\x87", 1, 0, -1);
    put_tiny (7, 136, "\x88", 1, 0, -1);
    put_tiny (7, 160, "\x89", 1, 0, -1);
    put_tiny (7, 184, "\x8A", 1, 0, -1);
}

static void draw_solid_graph (double a)
{
    double n;
    int i, m, k;

    m = statistics.graph.count < 120 ? 0 : statistics.graph.count - 120;
    
    for (i = 0 ; i < 0 + 120 ; i += 1) {
	n = i < statistics.graph.count ?
	    statistics.graph.samples[(i + m) % 120] :
	    0;

	if (statistics.mode == 2 && statistics.submode == 1) {
	    n = atan (n) * 40 / atan(a);
	} else {
	    n *= 40 / a;
	}

	if (n < -40) {
	    n = -40;
	} else if (n > 40) {
	    n = 40;
	}
	    
	k = 5 - ((abs(n) - 1) >> 3);

	put_bar (6, 8 + i, 1, n, 0, 1);

	if (n <= 32) {
	    put_bar (k, 8 + i, 1, (k - 1) << 3, 0, 0);
	}
    }
}

int main(void)
{
    int i, j, k, oldtab = -1;
    
    /* Set clock to 16Mhz. */

    CLKPR = 0x80, CLKPR = 0;

    /* Configure the IO ports. */

    BUTTONS_DDR = (BUTTONS_GROUND_ON_PIN << BUTTONS_GROUND_BIT) | (1 << 3);
    BUTTONS_PORT = ~BUTTONS_DDR;

    /* Set up all relevant pins of the display instruction
       port as outputs. */
    
    CONTROL_DDR = (1 << CONTROL_DI_BIT) |
	(1 << CONTROL_RW_BIT) |
	(1 << CONTROL_EN_BIT) |
	(1 << CONTROL_CS1_BIT) |
	(1 << CONTROL_CS2_BIT) |
	(1 << CONTROL_CS3_BIT) |
	(1 << CONTROL_RESET_BIT);

    /* Enable the USART, configure baud rate and
       set the word size to 8 bits with 1 stop bit
       and no parity. */

    UBRR1H = (unsigned char)(BAUD_PRESCALE >> 8);
    UBRR1L = (unsigned char)BAUD_PRESCALE;
    UCSR1B |= (1 << RXEN1) | (1 << TXEN1) | (1 << RXCIE1);
    UCSR1C = (1 << UCSZ10) | (1 << UCSZ11);

    TCCR2A |= (1 << WGM21);
    TCCR2B |= (1 << CS20) | (1 << CS21) | (1 << CS22);
    OCR2A = 255;
    OCR2B = 255;
    TIMSK2 |= (1 << OCIE2A);
    
    /* Set up Timer0 for PWM. */
    
    INDICATOR_DDR |= (1 << INDICATOR_BIT);
    TCCR0A = (1 << COM0A1) | (1 << WGM00) | (1 << WGM02);
    TCCR0B = (1 << CS00);

    /* Set up Timer1 to capture the tacho signal. */
    
    TCCR1A = 0;
    TCCR1B = (1 << ICNC1) | (1 << ICES1) | (1 << CS11);
    TIMSK1 = (1 << ICIE1) | (1 << TOIE1);

    /* Set up Timer3 with a 1 Hz cycle for
       flashing things on and off. */
    
    TCCR3B |= (1 << CS30) | (1 << CS32) | (1 << WGM32);
    OCR3A = F_CPU / 1024;
    OCR3B = F_CPU / 1024 / 5;
    TIMSK3 |= (1 << OCIE3A);
    
    /* Enable the ADC and set a prescaler of 128. */
    
    ADCSRA |= (1 << ADEN);
    ADCSRA |= (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0);

    /* Enable external interrupt INT1. */

    EICRA |= (1 << ISC10) | (1 << ISC11);
    EIMSK |= (1 << SIGNAL_INTERRUPT);

    /* Enable interrupts. */
    
    sei();
    
    /* Reset the display. */

    execute (1, 0, 0b00111111);
    execute (2, 0, 0b00111111);
    execute (3, 0, 0b00111111);
    
    load_waypoints ();
    load_odometers ();

    while (1) {
	char buffer[32];

	check_button_state();
	update_gps_data();
	update_temperature();
	update_engine_state();
	adjust_brightness();

	/* Blink the user led when the receiver
	   has no fix. */

	if (timing.blink && !gps.fix) {
	    BUTTONS_PORT |= (1 << 3);
	} else {
	    BUTTONS_PORT &= ~(1 << 3);
	}

	/* Clear the screen if content changed. */

	if (oldtab != display.tab) {
	    oldtab = display.tab;
	    
	    for (j = 1 ; j < 4 ; j += 1) {
		for (i = 0 ; i < 8 ; i += 1) {
		    set_address (j, i, 0);
		
		    for (k = 0 ; k < 64 ; k += 1) {
			write_column (j, 0);
		    }
		}
	    }
	}

	/* Draw the current tab. */
	
	if (display.tab < 5) {
	    put_tiny (1, 0,
		      "\x8B\x8B\x8B\x8B\x8B\x8B\x8E\x8B\x8B\x8B\x8B\x8B"
		      "\x8B\x8B\x8B\x8B\x8B\x8B\x8B\x8B",
		      0, 0, -1);
	    put_tiny (2, 69, "\x8D", 0, 0, -1);
	    put_tiny (3, 69, "\x8D", 0, 0, -1);
	    put_tiny (4, 69, "\x8D", 0, 0, -1);
	    put_tiny (5, 69,
		      "\x8C\x8B\x8B\x8B\x8B\x8B\x8B\x8B\x8B\x8B\x8B"
		      "\x8B\x8B\x8B",
		      0, 0, -1);
	
	    draw_bar_graph ();
	} else if (display.tab == 6) {
	    put_tiny (1, 0,
		      "\x8B\x8B\x8B\x8B\x8B\x8B\x8B\x8B\x8B\x8B\x8B\x8B"
		      "\x8B\x8B\x8B\x8B\x8B\x8B\x8E\x8B",
		      0, 0, -1);
	    put_tiny (2, 189, "\x8D", 0, 0, -1);
	    put_tiny (3, 189, "\x8D", 0, 0, -1);
	    put_tiny (4, 189, "\x8D", 0, 0, -1);
	    put_tiny (5, 189, "\x8D", 0, 0, -1);
	    put_tiny (6, 189, "\x8D", 0, 0, -1);
	    put_tiny (7, 189, "\x8D", 0, 0, -1);	    
	} else {
	    put_tiny (1, 0,
		      "\x8B\x8B\x8B\x8B\x8B\x8B\x8B\x8B\x8B\x8B\x8B\x8B"
		      "\x8E\x8B\x8B\x8B\x8B\x8B\x8B\x8B",
		      0, 0, -1);
	    put_tiny (2, 129, "\x8D", 0, 0, -1);
	    put_tiny (3, 129, "\x8D", 0, 0, -1);
	    put_tiny (4, 129, "\x8D", 0, 0, -1);
	    put_tiny (5, 129, "\x8D", 0, 0, -1);
	    put_tiny (6, 129, "\x8D", 0, 0, -1);
	    put_tiny (7, 129, "\x8D", 0, 0, -1);

	    put_tiny (2, 0, "\x8F", 1, 0, -1);
	    put_tiny (3, 0, "\x8F", 1, 0, -1);
	    put_tiny (4, 0, "\x8F", 1, 0, -1);
	    put_tiny (5, 0, "\x8F", 1, 0, -1);
	    put_tiny (6, 0, "\x8F", 1, 0, -1);
	    put_tiny (7, 0, "\x90\x91\x91\x91\x91\x91\x91\x91\x91\x91\x91\x91\x91", 0, 0, -1);

	    draw_solid_graph (statistics.resolutions[1] * 10);
	}

	if (display.tab == 3) {
	    /* Draw turn. */
	    
	    snprintf (buffer, 32, "%03u", abs(navigation.turn));
	    put_large (2, 61 - measure_large (buffer, 0), buffer, 0);
	    
	    snprintf (buffer, 32, "%c", navigation.turn < 0 ? 'L' : 'R');
	    
	    put_tiny (2, 62, "o", 0, 0, -1);
	    put_tiny (4, 62, buffer, 0, 2, -1);
	    
	    clear_tiny (4,
			62 + measure_tiny(buffer, 0),
			69 - 62 - measure_tiny(buffer, 0));
	    clear_tiny (5,
			62 + measure_tiny(buffer, 0),
			69 - 62 - measure_tiny(buffer, 0));
	} else if (display.tab == 4) {
	    /* Draw course over ground. */
	    
	    snprintf (buffer, 32, "%03u", (unsigned int)gps.course);
	    put_large (2, 61 - measure_large (buffer, 0), buffer, 0);

	    put_tiny (2, 62, "o", 0, 0, -1);

	    clear_tiny (4, 62, 6);
	    clear_tiny (5, 62, 6);	    
	} else if (display.tab < 5)  {
	    /* Draw speed over ground. */
	
	    snprintf (buffer, 32, "%3.0f", gps.speed);
	    put_large (2, 61 - measure_large (buffer, 0), buffer, 0);

	    clear_tiny (2, 62, 6);
	    clear_tiny (4, 62, 6);
	    clear_tiny (5, 62, 6);
	}	    

	/* Draw the header. */
	
	{
	    char *labels[] = {"Engine", "Distance", "Time", "Navigation",
			      "Waypoints", "Statistics", "Receiver"};

	    if (display.focus == 0) {
		put_tiny (0, 0, "\x80", 1, 0, -1);
	    } else {
		clear_tiny (0, 0, 4);
	    }
	    
	    put_tiny (0, 5, labels[display.tab], 1, 0, -1);
	    clear_tiny (0, 5 + measure_tiny (labels[display.tab], 1),
			60 - measure_tiny (labels[display.tab], 1));
	}

	/* Draw the date. */
	
	if (display.focus == 1) {
	    put_tiny (0, 189, "\x7f", 1, 0, -1);
	} else {
	    clear_tiny (0, 189, 3);
	}

	if (display.tab < 1) {
	    unsigned int day, month, year, weekday, hour, minute;
	    char *days[] = {"Sat", "Sun", "Mon", "Tue", "Wed", "Thu", "Fri"};
	    char *months[] = {"---", "Jan", "Feb", "Mar", "Apr", "May", "Jun",
			      "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
	    
	    day = gps.date / 10000;
	    month = (gps.date / 100) % 100;
	    year = gps.date % 100;
	    hour = (long int)gps.time / 10000;
	    minute = ((long int)gps.time / 100) % 100;
	    
	    /* Calculate the day of the week. */
	    
	    {
		int m, K;

		if (month < 3) {
		    m = month + 12;
		    K = year - 1;
		} else {
		    m = month;
		    K = year;
		}
		
		weekday = (day + (m + 1) * 26 / 10 + K + K / 4 + 105) % 7;
	    }
	    
	    if (month > 12) {
		month = 0;
	    }
	
	    snprintf (buffer, 32, "%.f\x81""C  %s %s %d, %02ld:%02ld", temperature.ambient,
		      days[weekday], months[month], day,
		      (long)hour, (long)minute);
	    put_tiny (0, 188 - measure_tiny (buffer, 1),
		      buffer, 1, 0, strchr(buffer, ':') - buffer);
	    clear_tiny (0, 65, 188 - 65 - measure_tiny (buffer, 1));
	} else if (display.tab == 1) {
	    snprintf (buffer, 32, "%.1f / %.1f km",
		      distance.odometers[4],
		      distance.odometers[3]);
	    put_tiny (0, 188 - measure_tiny (buffer, 1),
		      buffer, 1, 0, -1);
	    clear_tiny (0, 65, 188 - 65 - measure_tiny (buffer, 1));
	} else if (display.tab == 2) {
	    snprintf (buffer, 32, "%u:%02u.%03u",
		      timing.minutes[3], timing.seconds[3],
		      (unsigned)((double)timing.ticks[3] / 15625 * 1000));
	    put_tiny (0, 188 - measure_tiny (buffer, 1), buffer, 1, 0, -1);
	    clear_tiny (0, 65, 188 - 65 - measure_tiny (buffer, 1));
	} else if (display.tab > 2 && display.tab < 5) {
	    if (!strcmp (navigation.waypoint->name, ".")) {
		strcpy (buffer, "[Current loc.]");
	    } else if (strlen (navigation.waypoint->name) > 0) {
		strcpy (buffer, navigation.waypoint->name);
	    } else {
		strcpy (buffer, "[Hold to del.]");
	    }

	    put_tiny (0, 188 - measure_tiny (buffer, 1), buffer, 1, 0, -1);
	    clear_tiny (0, 65, 188 - 65 - measure_tiny (buffer, 1));
	} else if (display.tab == 5) {
	    if (statistics.mode == 0) {
		strcpy (buffer, "Speed(kmh)");
	    } else if (statistics.mode == 1) {
		strcpy (buffer, "Acceleration(g)");
	    } else if (statistics.mode == 2) {
		strcpy (buffer, "Lateral");
		snprintf (buffer, 32, "Lateral(%s)",
			  statistics.submode ? "deg" : "g");
	    }

	    put_tiny (0, 188 - measure_tiny (buffer, 1), buffer, 1, 0, -1);
	    clear_tiny (0, 65, 188 - 65 - measure_tiny (buffer, 1));
	} else if (display.tab == 6) {
	    /* Draw signal strength. */
	    
	    for (i = 0 ; i < 12 ; i += 1) {
		put_bar(0, 140 + 4 * i, 3, (i + 2) / 2 + 1, 1 - (i % 2), i < gps.satellites ? 1 : 0);
	    }

	    clear_tiny (0, 65, 188 - 65 - 48);
	}
	
	if (display.tab == 0) {
	    {
		static unsigned int latched;
		static char blink;

		if (blink != timing.blink) {
		    latched = engine.speed;
		    blink = timing.blink;
		}
	    
		/* Now the revs per minute. */

		put_tiny (2, 75, "Speed:", 1, 0, -1);
		snprintf (buffer, 32, "%u rpm", latched);
		put_tiny (2, 192 - measure_tiny (buffer, 1), buffer, 1, 0, -1);
		clear_tiny (2, 104, 192 - 104 - measure_tiny (buffer, 1));
	    }

	    /* And the temperature. */
    
	    put_tiny (3, 75, "Coolant:", 1, 0, -1);
	    if (temperature.coolant > 0) {
		snprintf (buffer, 32, "%.0f\x81""C", temperature.coolant);
	    } else {
		snprintf (buffer, 32, "--\x81""C");
	    }
	    put_tiny (3, 192 - measure_tiny (buffer, 1), buffer, 1, 0, -1);
	    clear_tiny (3, 111, 192 - 111 - measure_tiny (buffer, 1));

	    /* And the temperatures. */
    
	    put_tiny (4, 75, "Battery:", 1, 0, -1);
	    snprintf (buffer, 32, "%.2f V", engine.voltage);
	    put_tiny (4, 192 - measure_tiny (buffer, 1), buffer, 1, 0, -1);
	    clear_tiny (4, 119, 192 - 119 - measure_tiny (buffer, 1));
	} else if (display.tab == 1) {
	    for (i = 0 ; i < 3 ; i += 1) {
		if (display.focus == i + 2) {
		    put_tiny (2 + i, 71, "\x80", 1, 0, -1);
		} else {
		    clear_tiny (2 + i, 71, 4);
		}

		snprintf (buffer, 32, "%c:", 'A' + i);
		put_tiny (2 + i, 75, buffer, 1, 0,
			  (distance.running[i] || timing.blink) - 1);
	    
		snprintf (buffer, 32, "%.1f km", distance.odometers[i]);
		put_tiny (2 + i, 192 - measure_tiny (buffer, 1), buffer, 1, 0, -1);
		clear_tiny (2 + i, 83, 192 - 83 - measure_tiny (buffer, 1));
	    }
	} else if (display.tab == 2) {
	    for (i = 0 ; i < 3 ; i += 1) {
		if (display.focus == i + 2) {
		    put_tiny (2 + i, 71, "\x80", 1, 0, -1);
		} else {
		    clear_tiny (2 + i, 71, 4);
		}
		
		snprintf (buffer, 32, "%c:", 'A' + i);
		put_tiny (2 + i, 75, buffer, 1, 0,
			  (timing.running[i] || timing.blink) - 1);

		snprintf (buffer, 32, "%u:%02u.%03u",
			  timing.minutes[i], timing.seconds[i],
			  (unsigned)((double)timing.ticks[i] / 15625 * 1000));
		put_tiny (2 + i, 192 - measure_tiny (buffer, 1), buffer, 1, 0, -1);
		clear_tiny (2 + i, 83, 192 - 83 - measure_tiny (buffer, 1));
	    }
	} else if (display.tab == 3) {
	    /* Course over ground. */
	    
	    put_tiny (2, 75, "Track:", 1, 0, -1);
	    snprintf (buffer, 32, "%03d\x81", (int)gps.course);
	    
	    put_tiny (2, 192 - measure_tiny (buffer, 1), buffer, 1, 0, -1);
	    clear_tiny (2, 101, 192 - 101 - measure_tiny (buffer, 1));

	    /* Course to steer. */
	    
	    put_tiny (3, 75, "Bearing:", 1, 0, -1);
	    snprintf (buffer, 32, "%03d\x81", navigation.cts);
	    
	    put_tiny (3, 192 - measure_tiny (buffer, 1), buffer, 1, 0, -1);
	    clear_tiny (3, 111, 192 - 111 - measure_tiny (buffer, 1));
	    
	    /* Distance to waypoint. */

	    put_tiny (4, 75, "Distance:", 1, 0, -1);

	    if (navigation.distance < 1) {
		snprintf (buffer, 32, "%d m",
			  (int)(navigation.distance * 1000));
	    } else if (navigation.distance < 1000) {
		snprintf (buffer, 32, "%.1f km", navigation.distance);
	    } else {
		snprintf (buffer, 32, "%.0f km", navigation.distance);
	    }
	    
	    put_tiny (4, 192 - measure_tiny (buffer, 1), buffer, 1, 0, -1);
	    clear_tiny (4, 116, 192 - 116 - measure_tiny (buffer, 1));
	} else if (display.tab == 4) {
	    int blink;

	    for (i = 0 ; i < 2 ; i += 1) {
		if (display.focus == i + 2) {
		    put_tiny (2 + i, 71, "\x80", 1, 0, -1);
		} else {
		    clear_tiny (2 + i, 71, 4);
		}
	    }

	    if (navigation.editing == 9) {
		blink = 12;
	    } else if (navigation.editing >= 3) {
		blink = navigation.editing + 1;
	    } else if (navigation.editing > 0) {
		blink = navigation.editing;
	    } else {
		blink = -1;
	    }
	    
	    /* Latitude. */
	    
	    put_tiny (2, 75, "Lat.:", 1, 0, -1);
	    if (navigation.editing && display.focus == 2) {
		snprintf (buffer, 32, "%03ld.%06ld\x81 %c",
			  labs(navigation.digits) / 1000000,
			  labs(navigation.digits) % 1000000,
			  navigation.digits > 0 ? 'N' : 'S');
	    } else {
		snprintf (buffer, 32, "%02.6f\x81 %c",
			  fabs(navigation.waypoint->latitude),
			  navigation.waypoint->latitude > 0 ? 'N' : 'S');
	    }
	    
	    put_tiny (2, 192 - measure_tiny (buffer, 1), buffer, 1, 0,
		      display.focus == 2 ? blink : -1);
	    clear_tiny (2, 93, 192 - 93 - measure_tiny (buffer, 1));

	    /* Longitude. */

	    put_tiny (3, 75, "Long.:", 1, 0, -1);
	    if (navigation.editing && display.focus == 3) {
		snprintf (buffer, 32, "%03ld.%06ld\x81 %c",
			  labs(navigation.digits) / 1000000,
			  labs(navigation.digits) % 1000000,
			  navigation.digits > 0 ? 'E' : 'W');
	    } else {
		snprintf (buffer, 32, "%02.6f\x81 %c",
			  fabs(navigation.waypoint->longitude),
			  navigation.waypoint->longitude > 0 ? 'E' : 'W');
	    }
	    
	    put_tiny (3, 192 - measure_tiny (buffer, 1), buffer, 1, 0,
		      display.focus == 3 ? blink : -1);
	    clear_tiny (3, 100, 192 - 100 - measure_tiny (buffer, 1));

	    /* And the altitude. */

	    if (!strcmp (navigation.waypoint->name, ".")) {
		put_tiny (4, 75, "Alt.:", 1, 0, -1);
		snprintf (buffer, 32, "%d m", (int)gps.altitude);
		put_tiny (4, 192 - measure_tiny (buffer, 1), buffer, 1, 0, -1);
		clear_tiny (4, 93, 192 - 93 - measure_tiny (buffer, 1));
	    } else {
		clear_tiny (4, 75, 192 - 75);
	    }
	} else if (display.tab == 5) {
	    double mean, max, min, R_y;
	    int i;

	    i = statistics.mode;

	    mean = statistics.means[i].sum / statistics.means[i].samples;
	    R_y = statistics.resolutions[1];
	    
	    if (i > 0) {
		max = statistics.peaks[2 * i];
		min = statistics.peaks[2 * i + 1];
	    } else {
		max = statistics.peaks[0];
		min = 0;
	    }

	    if (i == 2 && statistics.submode == 1) {
		mean = atan (mean) * 180 / M_PI;
		min = atan (min) * 180 / M_PI;
		max = atan (max) * 180 / M_PI;
		R_y = atan (10 * R_y) * 180 / M_PI / 10;
	    }
	    
	    /* Mean. */
	    
	    if (display.focus == 2) {
		put_tiny (2, 131, "\x80", 1, 0, -1);
	    } else {
		clear_tiny (2, 131, 4);
	    }	
	    
    	    put_tiny (2, 135, "Mean:", 1, 0, -1);
	    snprintf (buffer, 32, "%.2f", mean);

	    if (!statistics.hold[i] || timing.blink) {
		put_tiny (2, 192 - measure_tiny (buffer, 1), buffer, 1, 0, -1);
		clear_tiny (2, 158, 192 - 158 - measure_tiny (buffer, 1));
	    } else {
		clear_tiny (2, 158, 192 - 158);
	    }

	    /* Peak. */
	    
	    if (display.focus == 3) {
		put_tiny (3, 131, "\x80", 1, 0, -1);
	    } else {
		clear_tiny (3, 131, 4);
	    }	
	    
	    put_tiny (3, 135, "P+:", 1, 0, -1);
	    
	    snprintf (buffer, 32, "%.2f", max);
	    
	    put_tiny (3, 192 - measure_tiny (buffer, 1), buffer, 1, 0, -1);
	    clear_tiny (3, 155, 192 - 155 - measure_tiny (buffer, 1));
	    
	    if (display.focus == 4) {
		put_tiny (4, 131, "\x80", 1, 0, -1);
	    } else {
		clear_tiny (4, 131, 4);
	    }	

	    if (statistics.mode > 0) {
		put_tiny (4, 135, "P-:", 1, 0, -1);
		snprintf (buffer, 32, "%.2f", min);
	    
		put_tiny (4, 192 - measure_tiny (buffer, 1), buffer, 1, 0, -1);
		clear_tiny (4, 153, 192 - 153 - measure_tiny (buffer, 1));
	    } else {
		clear_tiny (4, 135, 192 - 135);
	    }

	    /* Resolution. */

	    
	    if (display.focus == 5) {
		put_tiny (6, 131, "\x80", 1, 0, -1);
	    } else {
		clear_tiny (6, 131, 4);
	    }	
	    
	    if (display.focus == 6) {
		put_tiny (7, 131, "\x80", 1, 0, -1);
	    } else {
		clear_tiny (7, 131, 4);
	    }	
	    
    	    put_tiny (6, 135, "Rx:", 1, 0, -1);
	    snprintf (buffer, 32, "%.2f", statistics.resolutions[0]);

	    put_tiny (6, 192 - measure_tiny (buffer, 1), buffer, 1, 0, -1);
	    clear_tiny (6, 158, 192 - 158 - measure_tiny (buffer, 1));
	    
    	    put_tiny (7, 135, "Ry:", 1, 0, -1);
	    snprintf (buffer, 32, "%.2f", R_y);

	    put_tiny (7, 192 - measure_tiny (buffer, 1), buffer, 1, 0, -1);
	    clear_tiny (7, 158, 192 - 158 - measure_tiny (buffer, 1));
	} else if (display.tab == 6) {
	    /* Satellites used. */

	    put_tiny (2, 5, "Sattelites:", 1, 0, -1);
	    snprintf (buffer, 32, "%d", gps.satellites);
	    put_tiny (2, 187 - measure_tiny (buffer, 1), buffer, 1, 0, -1);
	    clear_tiny (2, 51, 187 - 51 - measure_tiny (buffer, 1));

	    /* Fix info. */
	    
	    {
		char *fix[3];

		if (gps.fix == 0) {
		    fix[0] = "no fix";
		} else if (gps.fix == 1) {
		    fix[0] = "GPS";
		} else if (gps.fix == 2) {
		    fix[0] = "DGPS";
		} else {
		    fix[0] = "DR";
		}

		if (gps.mode[1] == 1) {
		    fix[1] = "";
		} else if (gps.mode[1] == 1) {
		    fix[1] = ", 2D";
		} else {
		    fix[1] = ", 3D";
		}
		    
		if (tolower (gps.mode[0]) == 'm') {
		    fix[2] = ", manual";
		} else {
		    fix[2] = ", auto";
		}
		
		put_tiny (3, 5, "Mode:", 1, 0, -1);
		snprintf (buffer, 32, "%s%s%s", fix[0], fix[1], fix[2]);
		put_tiny (3, 187 - measure_tiny (buffer, 1), buffer, 1, 0, -1);
		clear_tiny (3, 28, 187 - 28 - measure_tiny (buffer, 1));
	    }

	    /* HDOP. */
    
	    put_tiny (4, 5, "Horizontal DOP:", 1, 0, -1);
	    snprintf (buffer, 32, "%.1f", gps.hdop);
	    put_tiny (4, 187 - measure_tiny (buffer, 1), buffer, 1, 0, -1);
	    clear_tiny (4, 77, 187 - 77 - measure_tiny (buffer, 1));

	    /* VDOP. */
    
	    put_tiny (5, 5, "Vertical DOP:", 1, 0, -1);
	    snprintf (buffer, 32, "%.1f", gps.vdop);
	    put_tiny (5, 187 - measure_tiny (buffer, 1), buffer, 1, 0, -1);
	    clear_tiny (5, 65, 187 - 65 - measure_tiny (buffer, 1));

	    /* PDOP. */
    
	    put_tiny (6, 5, "Positional DOP:", 1, 0, -1);
	    snprintf (buffer, 32, "%.1f", gps.pdop);
	    put_tiny (6, 187 - measure_tiny (buffer, 1), buffer, 1, 0, -1);
	    clear_tiny (6, 75, 187 - 75 - measure_tiny (buffer, 1));
	}
   }

    return 0;
}
