/* Copyright (C) 2011 Papavasileiou Dimitris                             
 *                                                                      
 * This program is free software: you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation, either version 3 of the License, or    
 * (at your option) any later version.                                  
 *                                                                      
 * This program is distributed in the hope that it will be useful,      
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        
 * GNU General Public License for more details.                         
 *                                                                      
 * You should have received a copy of the GNU General Public License    
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <avr/eeprom.h>
#include <math.h>

#include "state.h"

#define BASE 1024
#define ENTRY(i) (BASE + (i) * sizeof(struct entry))
#define MAX 64

struct entry {
    char name[8];
    float latitude, longitude;
};

static int length, current;

static int compare (const void *a, const void *b)
{
    const struct waypoint *p = a, *q = b;

    return strcmp (p->name, q->name);
}

void load_waypoints(void)
{
    int i, j;

    /* Enabling the following piece of code will format
       the waypoint memory and initialize it to the
       specified values when the device boots up. Disable
       again to prevent this from happening all the time. */
    
#if 0
    {
	int n;

	struct entry zero;
	struct entry entries[] = {
	    {"HOME", 38.00889462454789, 23.79861831665039},
	};

	n = sizeof (entries) / sizeof (entries[0]);

	memset (&zero, 0, sizeof (struct entry));

	for (i = 0 ; i < MAX ; i += 1) {
	    if (i < n) {
		eeprom_write_block(&entries[i],
				   (void *)ENTRY(i),
				   sizeof (struct entry));
	    } else {
		eeprom_write_block(&zero,
				   (void *)ENTRY(i),
				   sizeof (struct entry));

	    }
	}
    }
#endif	

    if (waypoints) {
    	free (waypoints);
    	waypoints = NULL;
    }
    
    for (i = 0, j = -1, length = 0 ; i < MAX ; i += 1) {
	struct entry p;

	eeprom_read_block (&p, (const void *)ENTRY(i), sizeof (struct entry));

	if (strnlen (p.name, 8) > 0) {
	    length += 1;
	    
	    waypoints = realloc (waypoints, length * sizeof (struct waypoint));
	    memset (&waypoints[length - 1], 0, sizeof (struct waypoint));
	    
	    strncpy(waypoints[length - 1].name, p.name, 8);
	    waypoints[length - 1].longitude = p.longitude;
	    waypoints[length - 1].latitude = p.latitude;
	    waypoints[length - 1].index = i;
	} else if (j < 0) {
	    j = i;
	}
    }

    qsort (waypoints, length, sizeof (struct waypoint), compare);
    
    if (j >= 0) {
    	length += 1;
	
    	waypoints = realloc (waypoints, length * sizeof (struct waypoint));
	memset (&waypoints[length - 1], 0, sizeof (struct waypoint));

    	strcpy(waypoints[length - 1].name, ".");
    	waypoints[length - 1].longitude = 0;
    	waypoints[length - 1].latitude = 0;
    	waypoints[length - 1].index = j;

	current = length - 1;
    } else {
	current = 0;
    }
    
    navigation.waypoint = &waypoints[current];
}

void switch_waypoint(void)
{
    current = (current + 1) % length;
    navigation.waypoint = &waypoints[current];
}

void update_waypoint(void) {
    struct entry p;

    strncpy (p.name, waypoints[current].name, 8);
    
    p.longitude = waypoints[current].longitude;
    p.latitude = waypoints[current].latitude;
	
    /* Update the selected waypoint. */
	
    eeprom_write_block(&p,
    		       (void *)ENTRY(waypoints[current].index),
    		       sizeof (struct entry));
 
    if (current == length - 1 || strlen(p.name) == 0) {
	load_waypoints();
    }
}

void update_navigational_data (void)
{
    double d, phi_2, phi_1, cosphi_1, cosphi_2, dlambda;

    /* Update the "current location" waypoint. */
    
    waypoints[length - 1].latitude = gps.latitude;
    waypoints[length - 1].longitude = gps.longitude;

    /* Calculate distance based on the haversine formula. */
    
    phi_1 = gps.latitude * M_PI / 180;
    phi_2 = navigation.waypoint->latitude * M_PI / 180;
    dlambda = (navigation.waypoint->longitude - gps.longitude) * M_PI / 180;
    cosphi_1 = cos(phi_1);
    cosphi_2 = cos(phi_2);
    
    d = sin (0.5 * (phi_2 - phi_1)) * sin (0.5 * (phi_2 - phi_1)) +
	cosphi_1 * cosphi_2 * sin (0.5 * dlambda) * sin (0.5 * dlambda);
    
    navigation.distance = 2 * 6371 * atan2(sqrt (d), sqrt (1 - d));

    /* Calculate bearing and turn. */
    
    d = atan2(sin(dlambda) * cosphi_2,
	      cosphi_1 * sin(phi_2) - sin(phi_1) * cosphi_2 * cos (dlambda));

    navigation.cts = fmod(d / M_PI * 180 + 360, 360);

    d = navigation.cts - gps.course;

    if (d > 180) {
	navigation.turn = d - 360;
    } else if (d < -180) {
	navigation.turn = d + 360;
    } else {
	navigation.turn = d;
    }
}
