/* Copyright (C) 2011 Papavasileiou Dimitris                             
 *                                                                      
 * This program is free software: you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation, either version 3 of the License, or    
 * (at your option) any later version.                                  
 *                                                                      
 * This program is distributed in the hope that it will be useful,      
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        
 * GNU General Public License for more details.                         
 *                                                                      
 * You should have received a copy of the GNU General Public License    
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <util/crc16.h>

#include "state.h"
#include "config.h"

#define THERMOMETER_INPUT_MODE()  THERMOMETER_DDR &= ~(1 << THERMOMETER_BIT)
#define THERMOMETER_OUTPUT_MODE() THERMOMETER_DDR |= (1 << THERMOMETER_BIT)
#define THERMOMETER_LOW()         THERMOMETER_PORT &= ~(1 << THERMOMETER_BIT)
#define THERMOMETER_HIGH()        THERMOMETER_PORT |= (1 << THERMOMETER_BIT)

static char converting;

static void write_bit(uint8_t bit){
    /* Pull line low for 1us. */
    
    THERMOMETER_LOW();
    THERMOMETER_OUTPUT_MODE();
    _delay_us(1);
    
    /* To write 1, release the line. */

    if(bit) {
	THERMOMETER_INPUT_MODE();
    }
	  
    /* Wait for 60us and release the line. */

    _delay_us(65);
    THERMOMETER_INPUT_MODE();
}

static uint8_t read_bit(void){
    uint8_t bit;

    /* Pull the line low for 1us. */

    THERMOMETER_LOW();
    THERMOMETER_OUTPUT_MODE();
    _delay_us(1);

    /* Release line and wait for line
       to stabilise (<15us). */

    THERMOMETER_INPUT_MODE();
    _delay_us(12);

    /* Read the line. */

    bit = (THERMOMETER_PIN & (1 << THERMOMETER_BIT)) > 0;

    /* Wait for another 50us to finish the slot. */

    _delay_us(50);
    
    return bit;
}

static uint8_t read_byte(void)
{
    uint8_t byte = 0;
    int i;

    for (i = 0 ; i < 8 ; i += 1) {
	byte |= (read_bit () << i);
    }

    return byte;
}

static void write_byte(uint8_t byte)
{
    int i;

    for (i = 0 ; i < 8 ; i += 1) {
	write_bit (byte & (1 << i));
    }
}

static uint8_t reset (void)
{
    char present;

    /* Pull line low and wait for 480us. */
	
    THERMOMETER_LOW();
    THERMOMETER_OUTPUT_MODE();

    _delay_us(480);

    /* Release line and wait for 60uS. */
	
    THERMOMETER_INPUT_MODE();
    _delay_us(60);

    /* Store line value and wait for another 480uS. */
	
    present = !(THERMOMETER_PIN & (1 << THERMOMETER_BIT));
    _delay_us(420);

    return present;
}

void update_temperature (void)
{
    uint8_t scratchpad[9], crc = 0;
    int i;
    
    if (!converting) {
    	/* Start a measurement. */

    	reset();
    	write_byte (0xCC);
    	write_byte (0x44);

    	converting = 1;
    }
    
    if(read_bit()) {
	/* Read the scratchpad. */
	
	reset ();
	write_byte(0xCC);
	write_byte(0xBE);

	for (i = 0, crc = 0;
	     i < 9;
	     crc = _crc_ibutton_update (crc, scratchpad[i]), i += 1) {
	    scratchpad[i] = read_byte();
	}

	if (crc == 0) {
	    temperature.ambient = *((int16_t *)&scratchpad) * 0.0625;
	}

	converting = 0;
    }
}
